/*
// author : Sven Goethel - Stapenhorststrasse 35a - 33615 Bielefeld
*/

#ifndef _RTTI_H
  #define _RTTI_H

  #include <string.h>
  #include <stdlib.h>
  #include "interror.h"

  /*
  // The rtti (run time type information) module supports
  // an object orientated handling for ansi-c struct-objects.
  //
  // Deep-Recursive creation, copy, assignment and destroying through
  // user declared structures is know avaiable for ANSI-C
  // structures.
  // Now, you are able to create a object orientated code in ANSI-C !!!
  // Deep means, that you can do pointer referenced code,
  // which is an registered rtti-type, with the ...Rtti functions !!!
  //
  // Each user-declared structure MUST include an rtti-structure
  // at the beginning of the data-structure if it wants to be
  // registered in the rtti-module !
  //
  // The rtti-structure identifys the type of the structure.
  //
  // With the 'rtti * NewRttiType(size_t sz)' and the
  // 'void RegisterRttiFct(...)' functions the user
  // can register a new rtti-type and can register individual
  // methods (functions) for compare, create, copycreate,
  // assignment and destroying.
  // The referenced memory of the returned rtti pointer from
  // the 'NewRttiType' function is managed by the rtti-module !!!!
  // Be sure to save the pointer for your rtti-identification,
  // but NEVER free or modificate the memory -- NEVER !!!!!!!!!!!!
  //
  // All rtti-registered structures can use the CmpRtti, CreateRtti, ...
  // functions. These functions will use the registered individual
  // functions, or ( if not registered ) they will use the default
  // memcpy, memcmp, calloc functions.
  //
  // The rtti-structure at the beginning of the user-declared and registered
  // structure will be set and updated ba all of the ...Rtti functions
  // ( see above ). The individual functions do not have to
  // initialize the rtti-structure and should not to avoid errors...
  //
  // It is more error proved, if you create your rtti-structures
  // with the ...CreateRtti functions, because of the guaranteed
  // rtti-initialisation !!!
  //
  // Befor leaving the main-program, do not forget to destroy the
  // rtti allocated memory with the 'void ResetRtti(void)' function !!!
  */

  typedef struct rtti
  {
    int              type;
    size_t           sz;                /* flat size in bytes incl. rtti */
  } rtti;

  #define C_RTTI(a) ((rtti *)(a))

  /*
  // all rtti arguments and return values (pointer),
  // are assumed to point on a valid structure, which included
  // the rtti structure FIRST !!!!
  //
  // The user defined functions MUST implement the same
  // semantics as described below !!!!!
  */

  /* return non zero, if a is the same type as b and if a has the same
     components as b
     INDIVIDUAL FUNCTION POSSIBLE !!!!
  */
  int CmpRtti ( const rtti * a, const rtti *b);

  /* returns a valid pointer, which references a valid structure of
     the rtti described type.
     INDIVIDUAL FUNCTION POSSIBLE !!!!
  */
  rtti *CreateRtti ( const rtti * );

  /* returns a valid pointer, which references a valid structure of
     the rtti described type and copys the components of b to a.
     INDIVIDUAL FUNCTION POSSIBLE !!!!
  */
  rtti *CopyCreateRtti ( const rtti * a );

  /* 'a' and 'b' must be a valid pointer .
     if 'a' and 'b' differs, 'a' will be destroyed (DestroyRtti)
     and new created (CopyCreateRtti).
     The result is the old or new pointer for a !!!
     Always usage : ' a = AssignRtti(a,b) ' to save the new a pointer !!!!
     INDIVIDUAL FUNCTION NOT POSSIBLE --> NOT NEEDED !!!!
  */
  rtti *AssignRtti ( rtti * a, const rtti *b );

  /* free all the memory allocated for the structure
     INDIVIDUAL FUNCTION POSSIBLE !!!!
  */
  void DestroyRtti ( rtti * );

  /* all the important create & copy functions .. */
  typedef int    ( * T_FCmpRtti)        ( const rtti * , const rtti *);
  typedef rtti * ( * T_FCreateRtti)     ( void );
  typedef rtti * ( * T_FCopyCreateRtti) ( const rtti * );
  typedef void   ( * T_FDestroyRtti)    ( rtti * );
  typedef void   ( * T_FResetRtti)      ( void );

  /* register a new rtti-type,
     and get the global type-description 'rtti' !!!!
     ( do not lose rtti, and do not modify or free rtti !!! )
  */
  rtti * NewRttiType(size_t sz);


  /* register individual copy & assignment (etc..) functions
     for your registered rtti-type,
  */
  void RegisterRttiFct(const rtti *rtype, T_FCmpRtti, T_FCreateRtti,
                       T_FCopyCreateRtti, T_FDestroyRtti, T_FResetRtti
                      );

  /* check the rtti structure type for valid values.
	   if ok the result is true, otherwise false.
		 be sure to check the result.
  */
  typedef enum { RTTI_OK=0, RTTI_INCORRECT_STRUCTURE, RTTI_INCORRECT_TABLE,
	               RTTI_UNKNOWN_TYPE, RTTI_NO_MORE_MEMORY, RTTI_WRONG_ERR
               } rtti_result ;

  rtti_result CheckRtti(const rtti *type);

	const char *GetRttiErrorStr(rtti_result rr);

  /* un-register all rtti-types and free all rtti-allocated memory ....
     BE SURE TO DO THIS JUST BEFOR EXITING THE PROGRAM !!!!
     ( or at a well defined run-time point .... )
  */
  void ResetRtti(void);

#endif

